# pylint: skip-file
# ruff: noqa
import hashlib
import os
import random
from copy import deepcopy

import gnupg
import pytest
from pytest_shell import fs

from opentaskpy import exceptions
from opentaskpy.taskhandlers import transfer
from tests.fixtures.pgp import *  # noqa: F403, F401
from tests.fixtures.ssh_clients import *  # noqa: F403, F401

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

# Create a task definition
sftp_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

# Non existent source directory
sftp_task_non_existent_source_dir_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src/nonexistentdir",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}


sftp_task_definition_no_permissions = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/etc",
            "rename": {
                "pattern": ".*taskhandler.*\\.txt",
                "sub": "passwd",
            },
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

sftp_file_watch_task_no_error_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*nofileexists.*\\.txt",
        "fileWatch": {"timeout": 1},
        "error": False,
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
}

sftp_with_fin_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*taskhandler.fin.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "flags": {
                "fullPath": "/home/application/testFiles/dest/sftp_with_fin.fin",
            },
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

# PCA move
sftp_pca_move_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_move\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/home/application/testFiles/archive",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

sftp_pca_move_task_definition_2 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_move_2\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/home/application/testFiles/archive/",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

sftp_pca_delete_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_delete\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {"action": "delete"},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}


sftp_pca_invalid_move_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_move_3\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/home/application/testFiles/archive/pca_move_bad.txt",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

sftp_pca_rename_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_rename_1\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "rename",
            "destination": "/home/application/testFiles/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

sftp_pca_rename_many_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "pca_rename_many.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "rename",
            "destination": "/home/application/testFiles/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

# Proxy task definition
sftp_proxy_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*taskhandler.proxy\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "transferType": "proxy",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}


sftp_destination_file_rename = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": "dest_rename.*taskhandler.*\\.txt",
        "protocol": {"name": "sftp", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "rename": {
                "pattern": "t(askha)ndler",
                "sub": "T\\1NDLER",
            },
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
    ],
}

fail_invalid_protocol_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.21",
        "directory": "/home/application/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "nonexistent"},
    },
}


def test_remote_handler(setup_sftp_keys):
    # Validate that given a transfer with sftp protocol, that we get a remote handler of type SFTP

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition)

    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "SFTPTransfer"

    # dest_remote_handler should be an array
    assert isinstance(transfer_obj.dest_remote_handlers, list)
    assert len(transfer_obj.dest_remote_handlers) == 1
    #  of SFTPTransfer objects
    assert transfer_obj.dest_remote_handlers[0].__class__.__name__ == "SFTPTransfer"


def test_sftp_basic(root_dir, setup_sftp_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    sftp_task_definition_copy = deepcopy(sftp_task_definition)
    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/test.taskhandler.txt")

    # Generate a random number
    random_number = random.randint(1, 1000000)
    # Test transferring to a subdir that doesn't exist, and creating it
    sftp_task_definition_copy["destination"][0][
        "directory"
    ] = f"/home/application/testFiles/dest/{random_number}"

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a false status, as we've not asked the directory to be created
    # Expect a RemoteTransferError
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()

    assert not os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt"
    )

    # Now run again, but ask for the dir to be created

    sftp_task_definition_copy["destination"][0]["createDirectoryIfNotExists"] = True

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt"
    )

    # Now run again, but set supportsPosixRename to false in the protocol definition
    sftp_task_definition_copy["destination"][0]["protocol"][
        "supportsPosixRename"
    ] = False

    # Delete the destination file
    os.remove(f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt")

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt"
    )

    # Finally run again, but set supportsStatAfterUpload to false in the protocol definition
    sftp_task_definition_copy["destination"][0]["protocol"][
        "supportsStatAfterUpload"
    ] = False

    # Delete the destination file
    os.remove(f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt")

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/{random_number}/test.taskhandler.txt"
    )

    # Change the destination directory to / and validate that it works
    sftp_task_definition_copy["destination"][0]["directory"] = "/"

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition_copy)

    # Run the transfer and expect a false status, because it cannot write to / on the remote
    # Expect a RemoteTransferError
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()


def test_sftp_basic_key_from_protocol_definition(root_dir, sftp_key_file):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    sftp_task_definition["source"]["protocol"]["credentials"]["key"] = sftp_key_file

    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/test.taskhandler.txt")


def test_sftp_basic_non_existent_source_directory(root_dir, setup_sftp_keys):
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None,
        "sftp-basic-non-existent-directory",
        sftp_task_non_existent_source_dir_definition,
    )

    # Run the transfer and expect a true status
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()


def test_sftp_basic_no_permissions(root_dir, setup_sftp_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "sftp-basic-no-permissions", sftp_task_definition_no_permissions
    )

    # Run the transfer and expect a true status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()


def test_sftp_filewatch_no_error(setup_sftp_keys):
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-no-file-no-error", sftp_file_watch_task_no_error_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_sftp_basic_write_fin(root_dir, setup_sftp_keys):
    # Delete any fin files that exist
    for file in os.listdir(f"{root_dir}/testFiles/sftp_2/dest"):
        if file.endswith(".fin"):
            os.remove(f"{root_dir}/testFiles/sftp_2/dest/{file}")

    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/test.taskhandler.fin.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_with_fin_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/test.taskhandler.fin.txt")

    # Check the fin file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/sftp_with_fin.fin")


def test_sftp_decrypt_incoming_file(
    tmpdir, root_dir, setup_sftp_keys, private_key, public_key
):

    # Create a test file
    source_file = f"{root_dir}/testFiles/sftp_1/src/test.decryption.txt"
    fs.create_files([{f"{source_file}": {"content": "test12345678"}}])

    # Checksum the file
    with open(f"{source_file}", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(public_key)

    # Encrypt the file
    with open(f"{source_file}", "rb") as f:
        status = gpg.encrypt_file(
            f,
            always_trust=True,
            recipients="test@example.com",
            output=f"{source_file}.gpg",
        )

        # Check status returns a 0 exit code
        assert status.ok

    # Check the output file exists
    assert os.path.exists(f"{source_file}.gpg")

    # Now we have an encrypted file (we can pretend we are collecting from elsewhere),
    # run a transfer to copy the file locally, and decrypt it
    sftp_task_definition_copy = deepcopy(sftp_task_definition)
    sftp_task_definition_copy["source"]["fileRegex"] = "test.decryption.txt.gpg"
    sftp_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": private_key,
    }

    # Override the destination
    sftp_task_definition_copy["destination"] = [
        {
            "directory": f"{tmpdir}/dest",
            "createDirectoryIfNotExists": True,
            "protocol": {"name": "local"},
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-decrypt", sftp_task_definition_copy)
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{tmpdir}/dest/test.decryption.txt")
    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{tmpdir}/dest/test.decryption.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum

    # Make sure only 1 file exists under the destination
    assert len(os.listdir(f"{tmpdir}/dest")) == 1


def test_sftp_encrypt_outgoing_file(
    tmpdir, root_dir, setup_sftp_keys, private_key, public_key
):

    source_file = f"{root_dir}/testFiles/sftp_1/src/test.encryption.txt"

    # Create a test file
    fs.create_files([{f"{source_file}": {"content": "test12345678"}}])

    # Checksum the file
    with open(f"{source_file}", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(private_key)

    # run a transfer to copy the file locally, and encrypt it
    sftp_task_definition_copy = deepcopy(sftp_task_definition)
    sftp_task_definition_copy["source"][
        "directory"
    ] = f"/home/application/testFiles/src"
    sftp_task_definition_copy["source"]["fileRegex"] = "test.encryption.txt"

    # Override the destination
    sftp_task_definition_copy["destination"] = [
        {
            "directory": f"{tmpdir}",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "public_key": public_key,
            },
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "sftp-encrypt", sftp_task_definition_copy)
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{tmpdir}/test.encryption.txt.gpg")

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{tmpdir}/test.encryption.txt.gpg", "rb"),
        output=f"{tmpdir}/test.encryption.txt",
    )

    assert decryption_data.ok
    assert decryption_data.returncode == 0

    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{tmpdir}/test.encryption.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum


def test_sftp_to_sftp_decrypt(
    root_dir, setup_sftp_keys, tmpdir, private_key, public_key
):
    # This is the same as the above test, but instead of the destination being local
    # we will transfer to sftp_2 instead

    # Create a test file
    source_file = f"{root_dir}/testFiles/sftp_1/src/test.decryption.e2e.txt"
    dest_file = f"{root_dir}/testFiles/sftp_2/dest/test.decryption.e2e.txt"
    fs.create_files([{f"{source_file}": {"content": "test12345678"}}])

    # Checksum the file
    with open(f"{source_file}", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(public_key)

    # Encrypt the file
    with open(f"{source_file}", "rb") as f:
        status = gpg.encrypt_file(
            f,
            always_trust=True,
            recipients="test@example.com",
            output=f"{source_file}.gpg",
        )

        # Check status returns a 0 exit code
        assert status.ok

    # Check the output file exists
    assert os.path.exists(f"{source_file}.gpg")

    # Now we have an encrypted file
    # run a transfer to copy the file to the other sftp, and decrypt it
    sftp_task_definition_copy = deepcopy(sftp_task_definition)
    sftp_task_definition_copy["source"]["fileRegex"] = "test.decryption.e2e.txt.gpg"
    sftp_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": private_key,
    }

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "sftp-decrypt", sftp_task_definition_copy)
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(dest_file)
    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(dest_file, "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum


def test_sftp_to_sftp_encrypt_outgoing_file(
    tmpdir, root_dir, setup_sftp_keys, private_key, public_key
):

    source_file = f"{root_dir}/testFiles/sftp_1/src/test.encryption.e2e.txt"

    # Create a test file
    fs.create_files([{f"{source_file}": {"content": "test12345678"}}])

    # Checksum the file
    with open(f"{source_file}", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(private_key)

    # run a transfer to copy the file locally, and encrypt it
    sftp_task_definition_copy = deepcopy(sftp_task_definition)
    sftp_task_definition_copy["source"][
        "directory"
    ] = f"/home/application/testFiles/src"
    sftp_task_definition_copy["source"]["fileRegex"] = "test.encryption.e2e.txt"

    # Override the destination
    sftp_task_definition_copy["destination"][0]["encryption"] = {
        "encrypt": True,
        "public_key": public_key,
    }

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "sftp-encrypt", sftp_task_definition_copy)
    assert transfer_obj.run()

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{root_dir}/testFiles/sftp_2/dest/test.encryption.e2e.txt.gpg", "rb"),
        output=f"{root_dir}/testFiles/sftp_2/dest/test.encryption.e2e.txt",
    )

    assert decryption_data.ok
    assert decryption_data.returncode == 0

    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{root_dir}/testFiles/sftp_2/dest/test.encryption.e2e.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum


def test_pca_move(root_dir, setup_sftp_keys):
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/sftp_1/archive"):
        os.remove(f"{root_dir}/testFiles/sftp_1/archive/{file}")

    # Create the test file
    fs.create_files(
        [{f"{root_dir}/testFiles/sftp_1/src/pca_move.txt": {"content": "test1234"}}]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move", sftp_pca_move_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/pca_move.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/src/pca_move.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/archive/pca_move.txt")

    # Create the next test file
    fs.create_files(
        [{f"{root_dir}/testFiles/sftp_1/src/pca_move_2.txt": {"content": "test1234"}}]
    )
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move-2", sftp_pca_move_task_definition_2
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/pca_move_2.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/src/pca_move_2.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/archive/pca_move_2.txt")

    fs.create_files(
        [{f"{root_dir}/testFiles/sftp_1/src/pca_move_3.txt": {"content": "test1234"}}]
    )
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move-invalid", sftp_pca_invalid_move_task_definition
    )

    # Run the transfer and expect a failure status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/pca_move_3.txt")
    # Check the source file should still exist
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/src/pca_move_3.txt")

    # Check the source file has not been archived
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/archive/pca_move_bad.txt")


def test_pca_delete(root_dir, setup_sftp_keys):
    # Create the next test file
    fs.create_files(
        [{f"{root_dir}/testFiles/sftp_1/src/pca_delete.txt": {"content": "test1234"}}]
    )
    transfer_obj = transfer.Transfer(
        None, "scp-pca-delete", sftp_pca_delete_task_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/pca_delete.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/src/pca_delete.txt")

    # Check the source file has been deleted
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/pca_delete.txt")


def test_destination_file_rename(root_dir, setup_sftp_keys):
    # Random number
    import random

    random_no = random.randint(1, 1000)

    # Create the test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/dest_rename_{random_no}_taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-dest-rename", sftp_destination_file_rename
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/dest_rename_{random_no}_TaskhaNDLER.txt"
    )


def test_pca_rename(root_dir, setup_sftp_keys):
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/sftp_1/archive"):
        os.remove(f"{root_dir}/testFiles/sftp_1/archive/{file}")

    # Create the test file
    fs.create_files(
        [{f"{root_dir}/testFiles/sftp_1/src/pca_rename_1.txt": {"content": "test1234"}}]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-rename", sftp_pca_rename_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/pca_rename_1.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/sftp_1/src/pca_rename_1.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/archive/pca_renamed_1.txt")


def test_pca_rename_many(root_dir, setup_sftp_keys):
    # Create the test file
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/sftp_1/archive"):
        os.remove(f"{root_dir}/testFiles/sftp_1/archive/{file}")

    # for 1 to 10
    for i in range(1, 10):
        fs.create_files(
            [
                {
                    f"{root_dir}/testFiles/sftp_1/src/pca_rename_many_{i}.txt": {
                        "content": "test1234"
                    }
                }
            ]
        )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-rename-name", sftp_pca_rename_many_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    for i in range(1, 10):
        assert os.path.exists(
            f"{root_dir}/testFiles/sftp_2/dest/pca_rename_many_{i}.txt"
        )
        # Check the source file no longer exists
        assert not os.path.exists(
            f"{root_dir}/testFiles/sftp_1/src/pca_rename_many_{i}.txt"
        )

        # Check the source file has been archived
        assert os.path.exists(
            f"{root_dir}/testFiles/sftp_1/archive/pca_renamed_many_{i}.txt"
        )


def test_sftp_proxy(root_dir, setup_sftp_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/test.taskhandler.proxy.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "sftp-basic", sftp_proxy_task_definition)
    local_staging_dir = transfer_obj.local_staging_dir

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_2/dest/test.taskhandler.proxy.txt"
    )

    # Ensure that local files are tidied up
    assert not os.path.exists(local_staging_dir)



"""Schemas for the configuration files."""

import copy
import importlib
import json
import sys
from importlib.resources import files
from pathlib import Path

from jsonschema import Draft202012Validator, validate, validators
from jsonschema.exceptions import ValidationError
from referencing import Registry, Resource

import opentaskpy.otflogging

TRANSFER_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "source": {
            "type": "object",
            "properties": {
                "protocol": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                    },
                    "required": ["name"],
                },
            },
            "required": ["protocol"],
        },
        "destination": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "protocol": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                        },
                        "required": ["name"],
                    },
                },
                "required": ["protocol"],
            },
        },
        "variables": {"type": "object"},
    },
    "required": ["type", "source"],
}

# Determine the type of transfer, and apply the correct sub schema based on the protocol used

EXECUTION_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "protocol": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
            },
            "required": ["name"],
        },
        "variables": {"type": "object"},
    },
    "required": ["type", "protocol"],
}

# Declare a constant that cannot be changed


BATCH_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "tasks": {
            "type": "array",
            "items": {"$ref": "http://localhost/batch/tasks.json"},
        },
    },
    "required": ["type", "tasks"],
}

logger = opentaskpy.otflogging.init_logging(__name__)

SCHEMAS = Path(str(files("opentaskpy.config").joinpath("schemas")))


def _extend_with_default(validator_class):  # type: ignore[no-untyped-def]
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):  # type: ignore[no-untyped-def]
        for _property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(_property, subschema["default"])

        yield from validate_properties(
            validator,
            properties,
            instance,
            schema,
        )

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


DefaultValidatingValidator = _extend_with_default(Draft202012Validator)  # type: ignore[no-untyped-call]


def _retrieve_from_filesystem(uri: str):  # type: ignore[no-untyped-def]
    if uri.startswith("http://localhost/"):
        path = SCHEMAS.joinpath(Path(uri.removeprefix("http://localhost/")))
    else:
        path = Path(uri.removeprefix("file://"))
    contents = json.loads(path.read_text(encoding="utf-8"))

    return Resource.from_contents(contents)


def validate_transfer_json(json_data: dict) -> bool:
    """Validate the JSON data against the transfer schema.

    Args:
        json_data (dict): The Transfer JSON definition

    Returns:
        bool: Whether the JSON data is valid or not
    """
    new_schema: dict

    try:
        validate(instance=json_data, schema=TRANSFER_SCHEMA)

        # If this works, then determine the protocol and apply the correct sub schema
        # Source protocol
        source_protocol = json_data["source"]["protocol"]["name"]

        # Dynamically apply the correct sub schema based on the protocol
        # We need to locate the schema files from the base location of this package
        # Get the directory of this package

        schema_dir = files("opentaskpy.config").joinpath("schemas")
        # This gets us the right path for default protocol schemas. But we need to be able to load custom schemas too
        # based on the protocol name, we need to find the location of the package using the files() function

        # Load the schema file for XXX_source
        resolver = Registry(retrieve=_retrieve_from_filesystem)

        # source_protocol is the class name within the plugin, we need to determine if it's a default protocol or a custom one
        # If it's a default protocol, then we can use the files() function to get the path to the schema file
        # If it's a custom protocol, then we need to look in the plugins directory for the schema file
        # Default protocols are the only ones that aren't prefixed with a package name, and we don't need to do anything more
        if "." in source_protocol:
            # Get the full package name from the class name (strip the class off the end)
            if (
                package_name := ".".join(source_protocol.split(".")[:-2])
            ) not in sys.modules:
                # Check the module is loadable
                importlib.import_module(package_name)
            # Get the path to the module
            module_path = files(package_name).joinpath("schemas")

            # Shorten the source_protocol to just the name of the class
            source_protocol = source_protocol.split(".")[-2]

            # Append new path to the resolver
            resolver.with_resource(module_path.as_uri(), module_path)  # type: ignore[attr-defined]
        else:
            # Default protocol
            module_path = schema_dir

        # Schema reference name should be schemas/transfer/xxx_source
        source_schema_name = Path(
            f"{module_path}/transfer/{source_protocol}_source.json"
        ).as_uri()

        # Destination protocol can be different for each destination in the array. Loop through each destination and get the protocol name
        schema_refs = []
        if "destination" in json_data:
            for destination in json_data["destination"]:
                # Get the protocol name
                destination_protocol = destination["protocol"]["name"]
                # As above, we need to determine the path of the protocol schema file if it's not a default protocol
                if "." in destination_protocol:
                    # Get the full package name from the class name (strip the class off the end)
                    package_name = ".".join(destination_protocol.split(".")[:-2])

                    if package_name not in sys.modules:
                        # Check the module is loadable
                        importlib.import_module(package_name)
                    # Get the path to the module
                    module_path = files(package_name).joinpath("schemas")

                    # Shorten the destination_protocol to just the name of the class
                    destination_protocol = destination_protocol.split(".")[-2]

                    # Append new path to the resolver
                    resolver.with_resource(module_path.as_uri(), module_path)  # type: ignore[attr-defined]
                else:
                    # Default protocol
                    module_path = schema_dir

                schema_def = {
                    "$ref": Path(
                        f"{module_path}/transfer/{destination_protocol}_destination.json"
                    ).as_uri()
                }

                # If schema_refs does not already contain the schema_def, then append it
                if schema_def not in schema_refs:
                    schema_refs.append(schema_def)

        # Update the TRANSFER_SCHEMA with the correct sub schema reference
        new_schema = copy.deepcopy(TRANSFER_SCHEMA)
        new_schema["properties"]["source"] = {"$ref": source_schema_name}

        # Now update the new_schema so that ["properties"]["destination"]["items"]["$ref"] = the list of destination protocols schema references
        # Schema reference name should be schemas/transfer/xxx_destination

        if "destination" in json_data and len(json_data["destination"]) > 0:
            new_schema["properties"]["destination"] = {
                "type": "array",
                "minItems": 0,
                "items": {
                    "oneOf": schema_refs,
                },
            }

        # Validate the new schema
        validator = DefaultValidatingValidator(
            new_schema,
            registry=resolver,
        )
        validator.validate(json_data)

    except ValidationError as err:
        print(err.message)  # noqa: T201
        return False
    return True


def validate_execution_json(json_data: dict) -> bool:
    """Validate the JSON data against the execution schema.

    Args:
        json_data (dict): The Execution JSON definition

    Returns:
        bool: Whether the JSON data is valid or not
    """
    try:
        validate(instance=json_data, schema=EXECUTION_SCHEMA)

        # If this works, then determine the protocol and apply the correct sub schema
        protocol = json_data["protocol"]["name"]

        # Dynamically apply the correct sub schema based on the protocol
        # We need to locate the schema files from the base location of this package
        # Get the directory of this package
        schema_dir = files("opentaskpy.config").joinpath("schemas")

        # Load the schema file for xxx
        resolver = Registry(retrieve=_retrieve_from_filesystem)

        if "." in protocol:
            # Get the full package name from the class name (strip the class off the end)
            if (package_name := ".".join(protocol.split(".")[:-2])) not in sys.modules:
                # Check the module is loadable
                importlib.import_module(package_name)
            # Get the path to the module
            module_path = files(package_name).joinpath("schemas")

            # Shorten the protocol to just the name of the class
            protocol = protocol.split(".")[-2]

            # Append new path to the resolver
            resolver.with_resource(module_path.as_uri(), module_path)  # type: ignore[attr-defined]
        else:
            # Default protocol
            module_path = schema_dir

        # Schema reference name should be schemas/execution/xxx
        schema_name = Path(
            f"{module_path}/execution/{protocol}/{protocol}.json"
        ).as_uri()

        # Update the schema with the correct sub schema reference
        new_schema = {}
        new_schema["$ref"] = schema_name

        # Validate the new schema
        # Validate the new schema
        validator = DefaultValidatingValidator(
            new_schema,
            registry=resolver,
        )
        validator.validate(json_data)

    except ValidationError as err:
        print(err.message)  # noqa: T201
        return False
    return True


def validate_batch_json(json_data: dict) -> bool:
    """Validate the JSON data against the batch schema.

    Args:
        json_data (dict): The Batch JSON definition

    Returns:
        bool: Whether the JSON data is valid or not
    """
    try:
        # Load the schema file for xxx
        resolver = Registry(retrieve=_retrieve_from_filesystem)

        validator = DefaultValidatingValidator(
            BATCH_SCHEMA,
            registry=resolver,
        )
        validator.validate(json_data)

    except ValidationError as err:
        print(err.message)  # noqa: T201
        return False
    return True



#!/bin/env python3
"""A simple script to do batch task validation."""

import argparse
import logging
import os
import sys

from opentaskpy.config.loader import ConfigLoader
from opentaskpy.otflogging import OTF_LOG_FORMAT

CONFIG_PATH = f"{os.getcwd()}/cfg"


def main(
    taskId: str | None = None,
    verbosity: int | None = None,
    configDir: str | None = None,
) -> bool:
    """Run the batch task validator.

    Args:
        taskId (str, optional): The task ID to validate. Defaults to None.
        verbosity (int, optional): The verbosity level. Defaults to None.
        configDir (str, optional): The config directory. Defaults to None.
    """
    # If given TaskId etc, then use those, otherwise use args parse
    if taskId:
        args = argparse.Namespace(
            taskId=taskId, verbosity=verbosity, configDir=configDir
        )
    else:
        parser = argparse.ArgumentParser()

        parser.add_argument(
            "-t",
            "--taskId",
            help="Name of the JSON config to run",
            type=str,
            required=True,
        )
        parser.add_argument(
            "-v",
            "--verbosity",
            help="Increase verbosity:\n3 - DEBUG\n2 - VERBOSE2\n1 - VERBOSE1",
            type=int,
        )
        parser.add_argument(
            "-c",
            "--configDir",
            help="Directory containing task configurations",
            type=str,
        )

        args = parser.parse_args()

    if args.configDir:
        global CONFIG_PATH  # pylint: disable=global-statement
        CONFIG_PATH = args.configDir

    logging_level = logging.INFO
    if args.verbosity == 3:
        logging_level = logging.DEBUG
    elif args.verbosity == 2:
        logging_level = 11
    elif args.verbosity == 1:
        logging_level = 12

    logging.addLevelName(11, "VERBOSE2")
    logging.addLevelName(12, "VERBOSE1")

    logging.basicConfig(
        format=OTF_LOG_FORMAT,
        level=logging_level,
        handlers=[logging.StreamHandler()],
    )

    # Set noop mode for other classes that expect it
    os.environ["OTF_NOOP"] = "true"

    logger = logging.getLogger()
    logger.setLevel(logging_level)

    logger = logging.getLogger(__name__)
    logger.log(11, f"Log verbosity: {args.verbosity}")

    # Create a config loader object
    config_loader = ConfigLoader(CONFIG_PATH)

    batch_task_definition = config_loader.load_task_definition(args.taskId)

    # Looking at the task definition, validate that any dependencies for tasks are defined as theur own task
    tasks = {}
    for task in batch_task_definition["tasks"]:
        task_definition = config_loader.load_task_definition(task["task_id"])

        order_id = task["order_id"]

        # Add it to the list of tasks
        tasks[order_id] = task_definition

    # Loop through the tasks and make sure the IDs are consecutive
    for i in range(1, len(tasks) + 1):
        if i not in tasks:
            logger.error(f"Task {i} is missing from the batch definition")
            return False

    # Loop through the tasks and ensure that the dependencies are valid
    for task in batch_task_definition["tasks"]:
        logger.debug(f"Checking dependencies for task {task['order_id']}")
        if "dependencies" not in task:
            continue
        for dependency in task["dependencies"]:
            if dependency not in tasks:
                logger.error(
                    f"Task {task['order_id']} has a dependency on task {dependency} which is not defined"
                )
                return False

    logger.info("Batch definition is valid")
    return True


if __name__ == "__main__":
    result = main()
    if not result:
        sys.exit(1)



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {
        "name": "ssh",
        "credentials": {
            "username": "test",
        },
    }


@pytest.fixture(scope="function")
def valid_protocol_definition_2():
    return {"name": "email"}


@pytest.fixture(scope="function")
def valid_source_definition(valid_protocol_definition):
    return {
        "hostname": "{{ HOST_A }}",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*\\.txt",
        "protocol": valid_protocol_definition,
    }


@pytest.fixture(scope="function")
def valid_destination_definition(valid_protocol_definition):
    return {
        "hostname": "somehost",
        "directory": "/tmp/testFiles/dest",
        "protocol": valid_protocol_definition,
    }


@pytest.fixture(scope="function")
def valid_destination_definition_2(valid_protocol_definition_2):
    return {
        "recipients": ["test@example.com"],
        "subject": "Here is your email",
        "protocol": valid_protocol_definition_2,
    }


def test_source_protocols(valid_source_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }

    # Remove protocol
    del json_data["source"]["protocol"]
    assert not validate_transfer_json(json_data)


def test_local_source_protocol(valid_source_definition):
    valid_source_definition["protocol"] = {"name": "local"}
    del valid_source_definition["hostname"]
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }
    assert validate_transfer_json(json_data)


def test_dest_with_different_protocols(
    valid_source_definition,
    valid_destination_definition,
    valid_destination_definition_2,
):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }

    # Add a valid destination
    json_data["destination"].append(valid_destination_definition)
    # Add a second valid destination
    json_data["destination"].append(valid_destination_definition_2)

    assert validate_transfer_json(json_data)


def test_encryption_with_public_key(valid_source_definition):

    valid_source_definition["encryption"] = {"public_key": "blah", "encrypt": True}

    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }

    assert validate_transfer_json(json_data)

    # Set sign to true and dont add a private key, expect a failure
    valid_source_definition["encryption"]["sign"] = True
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }

    assert not validate_transfer_json(json_data)

    # Add a private key
    valid_source_definition["encryption"]["private_key"] = "blah"
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [],
    }

    assert validate_transfer_json(json_data)



"""SFTP Remote Handlers.

This module contains the SFTP remote handlers for transfers.
"""

import glob
import os
import re
import stat
import time
from io import StringIO
from shlex import quote

from paramiko import AutoAddPolicy, Channel, RSAKey, SFTPClient, SSHClient
from tenacity import retry, stop_after_attempt, wait_exponential

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler


class SFTPTransfer(RemoteTransferHandler):
    """SFTP Transfer Handler."""

    TASK_TYPE = "T"

    FILE_NAME_DELIMITER = "|||"

    sftp_client: SFTPClient | None = None
    log_watch_start_row = 0

    def __init__(self, spec: dict):
        """Initialise the SFTPTransfer handler.

        Args:
            spec (dict): The spec for the transfer. This is either the source, or the
            destination spec.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        # Handle default values
        if "createDirectoryIfNotExists" not in spec:
            spec["createDirectoryIfNotExists"] = False

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return False, as all transfers require transfers via the worker."""
        return False

    def connect(self, hostname: str) -> None:
        """Connect to the remote host.

        Args:
            hostname (str): The hostname to connect to.
            sftp_client (SFTPClient, optional): An existing SFTPClient to use. Defaults to None.
        """
        if self.sftp_client and isinstance(self.sftp_client, SFTPClient):
            channel = self.sftp_client.get_channel()
            if channel is not None and isinstance(channel, Channel) and channel.active:
                return

        client_kwargs = {
            "hostname": hostname,
            "port": (self.spec["protocol"].get("port", 22)),
            "username": self.spec["protocol"]["credentials"]["username"],
            "timeout": 3,
            "allow_agent": False,
        }

        # If a custom key is set via env vars, then set that
        if (
            os.environ.get("OTF_SSH_KEY")
            and os.path.exists(str(os.environ.get("OTF_SSH_KEY")))
        ) and "keyFile" not in self.spec["protocol"]["credentials"]:
            self.logger.info("Loading custom private SSH key from OTF_SSH_KEY env var")
            key = RSAKey.from_private_key_file(str(os.environ.get("OTF_SSH_KEY")))
            client_kwargs["pkey"] = key

        # If a specific key file has been defined, then use that
        elif "keyFile" in self.spec["protocol"]["credentials"]:
            self.logger.info("Using key file from task spec")
            client_kwargs["key_filename"] = self.spec["protocol"]["credentials"][
                "keyFile"
            ]

        # If a private key has been defined as a string, then use that instead
        elif "key" in self.spec["protocol"]["credentials"]:
            self.logger.info("Using private key from task spec")
            key = RSAKey.from_private_key(
                StringIO(self.spec["protocol"]["credentials"]["key"])
            )
            client_kwargs["pkey"] = key

        # If a password has been defined, then use that
        elif "password" in self.spec["protocol"]["credentials"]:
            client_kwargs["password"] = self.spec["protocol"]["credentials"]["password"]

        self.connect_with_retry(client_kwargs)

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
    )
    def connect_with_retry(self, client_kwargs: dict) -> SSHClient:
        """Connect to the remote host with retry.

        Args:
            client_kwargs (dict): The kwargs to use for the SSH client.

        Returns:
            SSHClient: The SSH client.
        """
        try:
            ssh_client = SSHClient()
            ssh_client.set_log_channel(
                f"{__name__}.{ self.spec['task_id']}.paramiko.transport"
            )
            ssh_client.set_missing_host_key_policy(AutoAddPolicy())
            self.logger.info(f"Connecting to {client_kwargs['hostname']}")

            # Set additional timeout options to match the standard timeout
            client_kwargs["banner_timeout"] = client_kwargs["timeout"]
            client_kwargs["auth_timeout"] = client_kwargs["timeout"]
            client_kwargs["channel_timeout"] = client_kwargs["timeout"]

            ssh_client.connect(**client_kwargs)
            self.sftp_client = ssh_client.open_sftp()

        except Exception as ex:
            self.logger.error(f"Unable to connect to {client_kwargs['hostname']}: {ex}")
            raise ex

        return ssh_client

    def tidy(self) -> None:
        """Tidy up the SFTP connection.

        Shut down the SFTP connection and remove the remote transfer script from the
        remote host.
        """
        # Close connection
        if self.sftp_client:
            self.logger.info(f"[{self.spec['hostname']}] Closing SFTP connection")
            self.sftp_client.get_channel().close()  # type: ignore[union-attr]

            # Wait until the channel is closed
            time.sleep(0.25)
            for _ in range(2):
                if not self.sftp_client.get_channel().closed:  # type: ignore[union-attr]
                    time.sleep(0.5)
                else:
                    break

            self.sftp_client.close()

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Return list of files that match the source definition.

        Args:
            directory (str, optional): The directory to search in. Defaults to None.
            file_pattern (str, optional): The file pattern to search for. Defaults to
            None.

        Returns:
            dict: A dict of files that match the source definition.
        """
        self.connect(self.spec["hostname"])
        if not directory:
            directory = str(self.spec["directory"])
        if not file_pattern:
            file_pattern = str(self.spec["fileRegex"])

        self.logger.info(
            f"Searching for files in {directory} with pattern {file_pattern}"
        )

        self.logger.log(
            12,
            (
                f"[{self.spec['hostname']}] Searching in {directory} for files with"
                f" pattern {file_pattern}"
            ),
        )
        remote_files: dict = {}
        # Check the remote directory exists
        try:
            self.sftp_client.stat(directory)  # type: ignore[union-attr]
        except FileNotFoundError:
            self.logger.error(
                f"[{self.spec['hostname']}] Directory {directory} does not exist"
            )
            return remote_files

        remote_file_list = self.sftp_client.listdir(directory)  # type: ignore[union-attr]
        for file in list(remote_file_list):
            if re.match(file_pattern, file):
                # Get the file attributes
                file_attr = self.sftp_client.lstat(f"{directory}/{file}")  # type: ignore[union-attr]
                self.logger.log(12, f"File attributes {file_attr}")
                remote_files[f"{directory}/{file}"] = {
                    "size": file_attr.st_size,
                    "modified_time": file_attr.st_mtime,
                }

        return remote_files

    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str
    ) -> int:
        """Pull files to the worker.

        This function is used when we need to download source files from the source
        server onto the worker. These are then later pushed to the destination server

        Args:
            files (list): A list of files to download.
            local_staging_directory (str): The local staging directory to download the
            files to.

        Returns:
            int: 0 if successful, 1 if not.
        """
        result = 0
        # Connect to the source
        self.connect(self.spec["hostname"])

        # Create the staging directory locally
        if not os.path.exists(local_staging_directory):
            os.makedirs(local_staging_directory)

        # Download the files via SFTP
        for file in files:
            self.logger.info(
                f"[LOCALHOST] Downloading file {file} to {local_staging_directory}"
            )
            file_name = os.path.basename(file)
            try:
                self.sftp_client.get(file, f"{local_staging_directory}/{file_name}")  # type: ignore[union-attr]
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(
                    f"[LOCALHOST] Unable to download file locally via SFTP: {ex}"
                )
                result = 1

        return result

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the destination server.

        This function is used when the source files have been downloaded locally and
        need to be uploaded to the destination server. This would be expected to be
        called against the remote handler for the destination server.

        Args:
            local_staging_directory (str): The local staging directory to upload the
            files from.
            file_list (dict, optional): A list of files to upload. Defaults to None.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Connect to the destination server
        self.connect(self.spec["hostname"])
        # Check that the SFTP client is connected and active
        if not isinstance(self.sftp_client, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        # For SFTP, there is no staging directory, the files just go straight to where
        # they should do.

        destination_directory = quote(self.spec["directory"])

        # Sanitize the destination directory
        destination_directory = quote(destination_directory)

        self.logger.info(f"[{self.spec['hostname']}] Validating destination dir")
        # Use the SFTP client to check if the destination directory exists on the server
        remote_dir_exists = False
        try:
            self.sftp_client.stat(destination_directory)
            remote_dir_exists = True
        except OSError:
            self.logger.info(
                f"[{self.spec['hostname']}] Destination dir does not exist:"
                f" {destination_directory}"
            )

        if not remote_dir_exists and not self.spec["createDirectoryIfNotExists"]:
            self.logger.error(
                f"[{self.spec['hostname']}] Destination dir does not exist:"
                f" {destination_directory}"
            )
            return 1

        if not remote_dir_exists and self.spec["createDirectoryIfNotExists"]:
            self.logger.info(
                f"[{self.spec['hostname']}] Creating destination directory:"
                f" {destination_directory}"
            )
            self.sftp_client.mkdir(destination_directory)

        # Transfer the files
        result = 0

        if file_list:
            files = list(file_list.keys())
        else:
            # Get list of files in local_staging_directory
            files = glob.glob(f"{local_staging_directory}/*")
        for file in files:
            self.logger.info(f"[LOCALHOST] Transferring file via SFTP: {file}")
            file_name = os.path.basename(file)

            # Handle any rename that might be specified in the spec
            if "rename" in self.spec:
                rename_regex = self.spec["rename"]["pattern"]
                rename_sub = self.spec["rename"]["sub"]

                file_name = re.sub(rename_regex, rename_sub, file_name)
                self.logger.info(
                    f"[{self.spec['hostname']}] Renaming file to {file_name}"
                )

            mode = self.spec.get("mode", None)

            stat_after_upload = True
            if not self.spec["protocol"].get("supportsStatAfterUpload", True):
                stat_after_upload = False

            # If destination directory is the root, then set it to an empty string
            # so paths don't start with //
            if destination_directory == "/":
                destination_directory = ""

            try:

                # Check the protocol to see if supportsPosixRename is set
                if self.spec["protocol"].get("supportsPosixRename", True):

                    # While writing, the file should not have it's final name. Replace the
                    # file extension with .partial, and then rename it once the file has
                    # been transferred
                    file_name_partial = re.sub(r"\.[^.]+$", ".partial", file_name)

                    self.sftp_client.put(
                        file,
                        f"{destination_directory}/{file_name_partial}",
                        confirm=stat_after_upload,
                    )

                    # Rename the file to its final name
                    self.sftp_client.posix_rename(
                        f"{destination_directory}/{file_name_partial}",
                        f"{destination_directory}/{file_name}",
                    )
                else:
                    # Upload the file without using a temporary name
                    self.sftp_client.put(
                        file,
                        f"{destination_directory}/{file_name}",
                        confirm=stat_after_upload,
                    )

                if mode:
                    self.sftp_client.chmod(
                        f"{destination_directory}/{file_name}", int(mode, base=8)
                    )
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(
                    f"[{self.spec['hostname']}] Unable to transfer file via SFTP: {ex}"
                )
                result = 1

        return result

    def transfer_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list[str]) -> int:
        """Handle the post copy action specified in the config.

        Args:
            files (list[str]): A list of files that need to be handled.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])

        # Check that the SFTP client is connected and active
        if not isinstance(self.sftp_client, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        if self.spec["postCopyAction"]["action"] == "delete":
            # Loop through each file and use the sftp client to delete the files

            for file in files:
                try:
                    self.logger.info(f"Deleting file {file}")
                    self.sftp_client.remove(file)
                except OSError:
                    self.logger.error(
                        f"[{self.spec['hostname']}] Could not delete file {file} on"
                        " source host"
                    )
                    return 1

        if (
            self.spec["postCopyAction"]["action"] == "move"
            or self.spec["postCopyAction"]["action"] == "rename"
        ):
            # Use the SFTP client, and check that the destination directory exists
            move_dir = os.path.dirname(self.spec["postCopyAction"]["destination"])

            try:
                stat_result = self.sftp_client.stat(move_dir)
                # If it exists, then we need to ensure its a directory and not just a file
                if not stat.S_ISDIR(stat_result.st_mode):  # type: ignore[arg-type]
                    self.logger.error(
                        f"[{self.spec['hostname']}] Destination directory {move_dir} is"
                        " not a directory on source host"
                    )
                    return 1
            except OSError:
                self.logger.error(
                    f"[{self.spec['hostname']}] Destination directory {move_dir} does"
                    " not exist on source host"
                )
                return 1

                # Loop through the files and move them

            for file in files:
                try:
                    # If this is a move, then just move the file
                    if self.spec["postCopyAction"]["action"] == "move":
                        self.logger.info(
                            f"[{self.spec['hostname']}] Moving {file} to"
                            f" {self.spec['postCopyAction']['destination']}"
                        )
                        # Get the actual file name
                        file_name = os.path.basename(file)
                        self.sftp_client.posix_rename(
                            file,
                            f"{self.spec['postCopyAction']['destination']}/{file_name}",
                        )
                    # If this is a rename, then we need to rename the file
                    if self.spec["postCopyAction"]["action"] == "rename":
                        # Determine the new file name
                        new_file_dir = os.path.dirname(
                            self.spec["postCopyAction"]["destination"]
                        )
                        current_file_name = os.path.basename(file)

                        rename_regex = self.spec["postCopyAction"]["pattern"]
                        rename_sub = self.spec["postCopyAction"]["sub"]

                        new_file_name = re.sub(
                            rename_regex, rename_sub, current_file_name
                        )

                        self.logger.info(
                            f"[{self.spec['hostname']}] Renaming {file} to"
                            f" {new_file_dir}/{new_file_name}"
                        )
                        self.sftp_client.posix_rename(
                            file, f"{new_file_dir}/{new_file_name}"
                        )
                except OSError as e:
                    self.logger.error(f"[{self.spec['hostname']}] Error: {e}")
                    self.logger.error(
                        f"[{self.spec['hostname']}] Error moving or renaming file"
                        f" {file}"
                    )
                    return 1

        return 0

    def create_flag_files(self) -> int:
        """Create the flag files on the remote host.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])

        # Check that the SFTP client is connected and active
        if not isinstance(self.sftp_client, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        filename = self.spec["flags"]["fullPath"]

        try:
            # Use the SFTP client to create an empty file at this path
            self.sftp_client.file(filename, "w").close()

            # Set permissions on the file to whatever was specified in the spec,
            # otherwise we leave them as is
            # We cannot change ownership without using sudo, so we don't bother
            if "permissions" in self.spec:
                self.sftp_client.chmod(filename, int(self.spec["permissions"], base=8))

        except OSError as e:
            self.logger.error(f"[{self.spec['hostname']}] Error: {e}")
            self.logger.error(
                f"[{self.spec['hostname']}] Error creating flag file: {filename}"
            )
            return 1

        return 0



# pylint: skip-file
# ruff: noqa
import logging
import os
import random
import shutil
import subprocess

import pytest
from pytest_shell import fs

from tests.fixtures.ssh_clients import *  # noqa: F403

IMAGE_PREFIX = "opentaskpy_unittest"
MOVED_FILES_DIR = "archive"
DELIMITER = ","

"""
#################
Tests for running task-run via a docker container
#################
"""

# Get the current user and group id
user_id = os.getuid()
group_id = os.getgid()


@pytest.fixture(scope="module")
def image_name_dev():
    return f"{IMAGE_PREFIX}_dev_{random.randint(100000, 999999)}"


@pytest.fixture(scope="module")
def image_name():
    return f"{IMAGE_PREFIX}_{random.randint(100000, 999999)}"


@pytest.fixture(scope="module")
def tidy_images(root_dir):
    # Remove all images with the prefix
    result = subprocess.run(
        [
            "docker",
            "images",
            f"{IMAGE_PREFIX}_*",
        ],
        capture_output=True,
    )
    if result.stdout:
        images = result.stdout.decode("utf-8").split("\n")
        # Split the repository name from the rest of the line
        image_ids = [image.split(" ")[0] for image in images if image]
        for image_id in image_ids:
            # If this image name matches the prefix, remove it
            if image_id.startswith(IMAGE_PREFIX):
                logging.info(f"Removing image: {image_id}")
                result = subprocess.run(
                    [
                        "docker",
                        "rmi",
                        image_id,
                    ],
                    capture_output=True,
                )
                logging.info(result.stdout.decode("utf-8"))
                logging.info(result.stderr.decode("utf-8"))
    else:
        logging.info("No images to remove")


@pytest.fixture(scope="module")
def docker_build_dev_image(tidy_images, image_name_dev, root_dir):
    logging.info("Building dev docker image")
    logging.info(root_dir)

    # Trigger docker build
    command_args = [
        "docker",
        "build",
        "-t",
        image_name_dev,
        "-f",
        "Dockerfile.dev",
        f"{root_dir}/..",
    ]
    result = subprocess.run(
        command_args,
        capture_output=True,
    )
    assert result.returncode == 0


@pytest.fixture(scope="module")
def docker_build_image(tidy_images, image_name, root_dir):
    logging.info("Building docker image")
    logging.info(root_dir)

    # Trigger docker build
    command_args = [
        "docker",
        "build",
        "-t",
        image_name,
        "-f",
        "Dockerfile",
        f"{root_dir}/..",
    ]
    result = subprocess.run(
        command_args,
        capture_output=True,
    )
    assert result.returncode == 0


@pytest.fixture(scope="module")
def test_network_id():
    # Get the webdevops/ssh container that's running, and determine the network it's attached to
    result = subprocess.run(
        [
            "docker",
            "ps",
            "--filter",
            "name=^ssh_1",
            "--format",
            "{{.Networks}}",
        ],
        capture_output=True,
    )
    assert result.returncode == 0

    # We need to extract the network name
    network = result.stdout.decode("utf-8").strip()

    # Get the network ID
    result = subprocess.run(
        [
            "docker",
            "network",
            "inspect",
            network,
            "--format",
            "{{.Id}}",
        ],
        capture_output=True,
    )
    assert result.returncode == 0
    return result.stdout.decode("utf-8").strip()


@pytest.fixture(scope="module")
def log_dir(root_dir):
    return f"{root_dir}/testLogs"


@pytest.fixture(scope="function")
def clear_logs(log_dir):
    # Delete the output log directory
    if os.path.exists(log_dir):
        shutil.rmtree(log_dir)

    # Create an empty directory for the logs
    os.makedirs(log_dir, exist_ok=True)
    # Check the directory exists
    assert os.path.exists(log_dir)


@pytest.fixture(scope="function")
def create_test_file(root_dir):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/text.txt": {"content": "test1234"}}]
    )


def test_docker_run(
    setup_ssh_keys,
    docker_build_dev_image,
    test_network_id,
    image_name_dev,
    root_dir,
    env_vars,
    clear_logs,
    create_test_file,
):
    # Create the testLogs directory
    log_dir = f"{root_dir}/testLogs"
    os.makedirs(log_dir, exist_ok=True)

    # Run the container
    logging.info("Running docker container")
    command_args = [
        "docker",
        "run",
        "--rm",
        "--user",
        f"{user_id}:{group_id}",
        "--network",
        test_network_id,
        "--volume",
        f"{root_dir}:/test",
        "--volume",
        f"{log_dir}:/logs",
        "--volume",
        "/tmp/variable_lookup.txt:/tmp/variable_lookup.txt",
        "-e",
        "OTF_SSH_KEY=/test/testFiles/id_rsa",
        image_name_dev,
        "task-run",
        "-t",
        "scp-basic",
        "-v",
        "2",
        "-c",
        "/test/cfg",
    ]

    result = subprocess.run(
        command_args,
        capture_output=True,
    )
    logging.info(result.stdout.decode("utf-8"))
    logging.info(result.stderr.decode("utf-8"))

    assert result.returncode == 0


def test_standard_docker_image(
    setup_ssh_keys,
    docker_build_image,
    test_network_id,
    image_name,
    root_dir,
    log_dir,
    env_vars,
    clear_logs,
    create_test_file,
):
    # Create a test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/text.txt": {"content": "test1234"}}]
    )

    # This image pulls down whatever is on pypi, so we're not really testing the code here. Just that we can call a simple transfer.
    # We want to check that the logging works correctly when running in a docker container and mapping volumes
    # Run the container
    logging.info("Running docker container")
    command_args = [
        "docker",
        "run",
        "--rm",
        "--user",
        f"{user_id}:{group_id}",
        "--network",
        test_network_id,
        "--volume",
        f"{root_dir}:/test",
        "--volume",
        "/tmp/variable_lookup.txt:/tmp/variable_lookup.txt",
        "--volume",
        f"{root_dir}/testLogs:/logs",
        "--env",
        "OTF_RUN_ID=docker_log_test",
        image_name,
        "task-run",
        "-t",
        "scp-basic",
        "-v",
        "2",
        "-c",
        "/test/cfg",
    ]
    logging.info(" ".join(command_args))

    subprocess.run(
        command_args,
        capture_output=True,
    )
    # We dont care whether this worked or not, we just want to check the logs
    # Check that the log file exists containing scp-basic in the name in log_dir
    log_files = os.listdir(f"{log_dir}/docker_log_test")
    assert len(log_files) == 1
    assert "scp-basic" in log_files[0]
